c*************************************************************************
c IO_WRITE_FILTER
c*************************************************************************
c accumulate array of orbital elements for test particles and planets,
c filter/decimate it and write out a whole frame to an real*4 binary file.
c
c             Input:
c                 time          ==>  current time (real scalar)
c                 nbod          ==>  number of massive bodies (int scalar)
c                 ntp           ==>  number of test particles (int scalar)
c                 mass          ==>  mass of bodies (real array)
c                 xh,yh,zh      ==>  current position in helio coord 
c                                    (real arrays)
c                 vxh,vyh,vzh   ==>  current velocity in helio coord 
c                                    (real arrays)
c                 xht,yht,zht    ==>  current part position in helio coord 
c                                      (real arrays)
c                 vxht,vyht,vzht ==>  current velocity in helio coord 
c                                        (real arrays)
c                 istat           ==>  status of the test paricles
c                                      (2d integer array)
c                                      istat(i,1) = 0 ==> active:  = 1 not
c                                      istat(i,2) = -1 ==> Danby did not work
c                 oname           ==> output file name (character string) 
c                 iu              ==> unit number to write to
c                 fopenstat       ==>  The status flag for the open 
c                                      statements of the output files.  
c                                          (character*80)
c
c Other filtration parameters in common block /filter/!
c
c Remarks: Based on io_write_frame
c Authors: Miroslav Broz, miroslav.broz@email.cz
c Date:    Dec 5th 1998
c Revision: Dec 8th 2003
c - fmft filter now correctly handles shift of p, q buffers
c Revision: May 12th 2003
c - output filtered elements e and inc are calculated from h,k,p,q
c (in accordance with proper elements filter, which uses h,k,p,q variables)
c Revision: Mar 15th 2002
c - now the filtering is done for a,h=e*sin(peri+node),k,p,q,e,inc
c - corrected startup skip of buffers due to alignment of output data
c - added OpenMP directives for parallelizing compilers
c Revision: Dec 2nd 2002
c - corrected output time (shifted by half of the initial buffer fillup)
c - flt_cb.inc include file with /filter/ common block
c Revision: Mar 18th 2003
c - write of mean elements might be switched-off
c Revision: Apr 8th 2003
c - output could be in real*8 precision

      subroutine io_write_filter(time,nbod,ntp,mass,xh,yh,zh,vxh,
     &           vyh,vzh,xht,yht,zht,vxht,vyht,vzht,istat,oname,
     &           iu,fopenstat)

      include 'swift.inc'
      include 'filter.inc'
      include 'flt_cb.inc'
      include 'oscel_cb.inc'

c...  Inputs: 
      integer nbod,ntp,iu
      real*8 mass(nbod),time
      integer istat(NTPMAX,NSTAT)
      real*8 xh(nbod),yh(nbod),zh(nbod)
      real*8 vxh(nbod),vyh(nbod),vzh(nbod)
      real*8 xht(ntp),yht(ntp),zht(ntp)
      real*8 vxht(ntp),vyht(ntp),vzht(ntp)
      character*80 oname,fopenstat
      integer iflgchk

c...  Internals
      integer i,n,id,k
      integer ialpha,ierr,FNskip,FNcnt

      real*8 gm,capa,e,inc,node,peri,capm,varpi,sininc2,tout
      integer i1st,i2nd,i3rd	! =0 first time through; =1 after
      integer FN(FMAXC)	! index in elmts() buffer, (1;FBUFN)
      save i1st,i2nd,i3rd,FN,FNskip,FNcnt
c
c  orbital elements are stored in common block /elements/, will be
c  used for computation of seasonal Yarkovsky force in getacc_tp.f;
c  dimensions of the array: i) width of the buffer, ii) 7 orbital
c  elements, iii) PLs and TPS, iv) number of buffer levels
c 
      real*8 elmts(FMAXN,7,-FMAXNPL:FMAXNTP,FMAXC)
      save elmts	! VERY IMPORTANT (otherwise omf77 crashes on SIGSEGV)

c  g77/RH60 compiler does NOT include this equivalence => assign idx=FN(1)
c  just before return from this subroutine
c      equivalence (idx,FN(1))
c  another solution -> pass only a small copy of elmts() array (see oscel_cb.inc)

      data i1st,i2nd,i3rd,FNcnt/0,0,0,0/

c----
c...  Executable code 

c  first time through
      if (i1st.eq.0) then
c  do this block only once
        if (i2nd.eq.0) then
          if (filter_write) then
            call io_open(iu,oname,fopenstat,'UNFORMATTED',ierr)
            if(ierr.ne.0) then
              write(*,*) ' SWIFT ERROR: in io_write_frame: '
              write(*,*) '     Could not open binary output file:'
              call util_exit(1)
            endif
            close(iu)
          endif

c  index for filter_N levels of the buffer
          do i = 1, filter_N
            FN(i) = 0
          enddo
c
c  Skip first data due to time-alignment in filtered output.
c
c  FNskip ... how many points should we skip?
          FNskip = (FNinibuf/FNdtout + 1) * FNdtout - FNinibuf
          i2nd = 1
        endif

c  wait until we can skip back to the beginning of the buffer
c  (FNcnt is the counter)
        if (FNcnt.lt.FNskip) then
          FNcnt = FNcnt + 1
          FN(1) = 0
        else
          i1st = 1
        endif
      endif

c  calculate orbital elements for planets and test particles
      t0 = time		! remember the epoch of mean anomaly
      upflg_osc = .true.	! update flag for io_write_proper.f subroutine
      FN(1)=FN(1)+1
      gm=mass(1)
!$OMP PARALLEL DO
!$OMP& PRIVATE (i, id, ialpha, capa, e, inc, node, peri, capm,
!$OMP&   varpi, sininc2)
!$OMP& SHARED (ntp, xht, yht, zht, vxht, vyht, vzht, gm,
!$OMP&   elmts, osc_elmts)
      do i = 1, ntp
        if(istat(i,1).eq.0) then
          id=i
          call orbel_xv2el(xht(i),yht(i),zht(i),vxht(i),vyht(i),vzht(i),
     &      gm,ialpha,capa,e,inc,node,peri,capm)
          elmts(FN(1),1,id,1) = capa
c  Sep 20th 2000: changed the angle in the definition of h, k from peri to varpi
          varpi = peri + node
          elmts(FN(1),2,id,1) = e * cos(varpi)
          elmts(FN(1),3,id,1) = e * sin(varpi)
          sininc2 = sin(0.5d0*inc)
          elmts(FN(1),4,id,1) = sininc2 * cos(node)
          elmts(FN(1),5,id,1) = sininc2 * sin(node)
c  Sep 20th 2000: separate filtering of e, inc
c          elmts(FN(1),6,id,1) = e
c          elmts(FN(1),7,id,1) = inc
          osc_elmts(1,id) = capa
          osc_elmts(2,id) = e
          osc_elmts(3,id) = inc
          osc_elmts(4,id) = peri
          osc_elmts(5,id) = node
          osc_elmts(6,id) = capm
        endif
      enddo
!$OMP END PARALLEL DO

!$OMP PARALLEL DO
!$OMP& PRIVATE (i, gm, id, ialpha, capa, e, inc, node, peri, capm,
!$OMP&   varpi, sininc2)
!$OMP& SHARED (nbod, xh, yh, zh, vxh, vyh, vzh, elmts)
      do i = 2, nbod
        gm=mass(1)+mass(i)
        id=-i
        call orbel_xv2el(xh(i),yh(i),zh(i),vxh(i),vyh(i),vzh(i),
     &    gm,ialpha,capa,e,inc,node,peri,capm)
        elmts(FN(1),1,id,1) = capa
        varpi = peri + node
        elmts(FN(1),2,id,1) = e * cos(varpi)
        elmts(FN(1),3,id,1) = e * sin(varpi)
        sininc2 = sin(0.5d0*inc)
        elmts(FN(1),4,id,1) = sininc2 * cos(node)
        elmts(FN(1),5,id,1) = sininc2 * sin(node)
c        elmts(FN(1),6,id,1) = e
c        elmts(FN(1),7,id,1) = inc
        osc_elmts(1,id) = capa
        osc_elmts(2,id) = e
        osc_elmts(3,id) = inc
        osc_elmts(4,id) = peri
        osc_elmts(5,id) = node
        osc_elmts(6,id) = capm
      enddo
!$OMP END PARALLEL DO

c
c  test if buffers are filled up
c
      do n=1,filter_N-1
        if (FN(n).ge.FBUFN) then

          FN(n+1)=FN(n+1)+1

c  filter the array
          if (filter_stat(n:n).eq.'A') then
            call filter_elmts(elmts,nbod,ntp,istat,FN,n,filter_M,A)
          else	! if (filter_stat(n:n).eq.'B') then
            call filter_elmts(elmts,nbod,ntp,istat,FN,n,filter_M,B)
          endif

c  decime with factor filter_D
          FN(n)=FN(n)-filter_D(n)

c  shift left
          call filter_shift(elmts,nbod,ntp,n,FBUFN,filter_D)
        endif
      enddo
c
c  test if the last buffer is filled up and write filtered elmts
c
      if (FN(filter_N).ge.FBUFN) then
        tout = time - FNinibuf*dtfilter/2.d0
        if (filter_write) then
          if (i3rd.eq.0) then
            call io_open(iu,oname,fopenstat,'UNFORMATTED',ierr)
            i3rd = 1
          else
            call io_open(iu,oname,'append','UNFORMATTED',ierr)
          endif
          if (filter_write_r8) then
            call io_write_hdr_r8(iu,tout,nbod,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,nbod,ntp,istat)
          endif
        endif
        n=filter_N
        if (filter_stat(n:n).eq.'A') then
          call filter_elmts_write(iu,elmts,n,nbod,ntp,istat,
     &      A,filter_M,filter_write,filter_write_r8)
        else	! if (filter_stat(n:n).eq.'B') then
          call filter_elmts_write(iu,elmts,n,nbod,ntp,istat,
     &      B,filter_M,filter_write,filter_write_r8)
        endif

c  "decime" with factor filter_D
        FN(n)=FN(n)-filter_D(n)

c  shift left
        call filter_shift(elmts,nbod,ntp,n,FBUFN,filter_D)
        if (filter_write) then
          close(iu)
        endif
      endif

      return
      end	! io_write_filter
c----------------------------------------------------------------------

c**********************************************************************
      subroutine filter_elmts(elmts,nbod,ntp,istat,FN,n,filter_M,d)
c**********************************************************************
c
c  filter the array elmts with filter d and store the filtered
c  orbital elements in the level n+1 in array elmts
c
      include 'swift.inc'
      include 'filter.inc'

      real*8 elmts(FMAXN,7,-FMAXNPL:FMAXNTP,FMAXC)
c  Mar 15th 2002: added istat parameter
      integer nbod,ntp,istat(NTPMAX,NSTAT),n,FN(FMAXC),filter_M
      real*8 d(-FMAXM:FMAXM)

c  functions
      real*8 filter_filter

      integer i,k,id

!$OMP PARALLEL
!$OMP& SHARED (ntp, elmts, FN, n, d, filter_M)
!$OMP DO
!$OMP& PRIVATE (i, k, id)
      do i=2,nbod
        id=-i
        do k=1,7
          elmts(FN(n+1),k,id,n+1)=
     &      filter_filter(d,elmts(1,k,id,n),filter_M)
        enddo
      enddo
!$OMP END DO NOWAIT

!$OMP DO
!$OMP& PRIVATE (i, k, id)
      do i=1,ntp
        if (istat(i,1).eq.0) then
          id=i
          do k=1,7
            elmts(FN(n+1),k,id,n+1)=
     &        filter_filter(d,elmts(1,k,id,n),filter_M)
          enddo
        endif
      enddo
!$OMP END DO
!$OMP END PARALLEL

      return
      end	! filter_elmts

c**********************************************************************
      subroutine filter_elmts_write(iu,elmts,n,nbod,ntp,istat,d,FM,
     :  filter_write,filter_write_r8)
c**********************************************************************
c
c  filter the array elmts with filter d and write filtered
c  orbital elements to the binary output file iu
c
      include 'swift.inc'
      include 'filter.inc'
      include 'meanel_cb.inc'

      integer iu,nbod,ntp,n,FM,istat(NTPMAX,NSTAT)
      real*8 elmts(FMAXN,7,-FMAXNPL:FMAXNTP,FMAXC)
      real*8 d(-FMAXM:FMAXM)
      logical filter_write,filter_write_r8
c temp vars
      integer i,id
      real*8 a,k,h,q,p,e,inc,peri,node,varpi
c  functions
      real*8 filter_filter, zero2pi

      if (filter_write) then	! write output-file?!

c  I did'nt parallelize these 2 cycles, because of i/o routines.
        do i=2,nbod
          id=-i
          a = filter_filter(d,elmts(1,1,id,n),FM)
          k = filter_filter(d,elmts(1,2,id,n),FM)
          h = filter_filter(d,elmts(1,3,id,n),FM)
          q = filter_filter(d,elmts(1,4,id,n),FM)
          p = filter_filter(d,elmts(1,5,id,n),FM)
          varpi = atan2(h,k)
          node = atan2(p,q)
          peri = zero2pi(varpi-node)
c          e = filter_filter(d,elmts(1,6,id,n),FM)
c          inc = filter_filter(d,elmts(1,7,id,n),FM)
c  May 12th 2003: calculate e, inc again from h,k,p,q
c  (Because proper elements are calculted from these and
c  we need proper elements to be reproducible!)
          e = sqrt(k*k+h*h)
          inc = 2.0d0 * asin(sqrt(q*q+p*p))

          if (filter_write_r8) then
            call io_write_line_r8(iu,id,a,e,inc,node,peri,0.0d0)
          else
            call io_write_line_r(iu,id,a,e,inc,node,peri,0.0d0)
          endif

c  do NOT store planetary mean elements also
c          mean_elmts(1,id) = a
c          mean_elmts(2,id) = k
c          mean_elmts(3,id) = h
c          mean_elmts(4,id) = q
c          mean_elmts(5,id) = p
c          mean_elmts(6,id) = e
c          mean_elmts(7,id) = inc
        enddo
      endif

      do i=1,ntp
        if (istat(i,1).eq.0) then
          id=i
          a = filter_filter(d,elmts(1,1,id,n),FM)
          k = filter_filter(d,elmts(1,2,id,n),FM)
          h = filter_filter(d,elmts(1,3,id,n),FM)
          q = filter_filter(d,elmts(1,4,id,n),FM)
          p = filter_filter(d,elmts(1,5,id,n),FM)
          varpi = atan2(h,k)
          node = atan2(p,q)
          peri = zero2pi(varpi-node)
c          e = filter_filter(d,elmts(1,6,id,n),FM)
c          inc = filter_filter(d,elmts(1,7,id,n),FM)
          e = sqrt(k*k+h*h)
          inc = 2.0d0 * asin(sqrt(q*q+p*p))

          if (filter_write) then
            if (filter_write_r8) then
              call io_write_line_r8(iu,id,a,e,inc,node,peri,0.0d0)
            else
              call io_write_line_r(iu,id,a,e,inc,node,peri,0.0d0)
            endif
          endif

c  store output mean elements in array mean_elmts (and update flag)
          mean_elmts(1,id) = a
          mean_elmts(2,id) = k
          mean_elmts(3,id) = h
          mean_elmts(4,id) = q
          mean_elmts(5,id) = p
          mean_elmts(6,id) = e
          mean_elmts(7,id) = inc
        endif
      enddo

      upflg_mean = .true.

      return
      end	! filter_elmts_write

c**********************************************************************
      subroutine filter_shift(elmts,nbod,ntp,n,FBUFN,filter_D)
c**********************************************************************
c
c  shift left elements in array elmts by filter_D(n) / decimation
c
      implicit none
      include 'filter.inc'

      real*8 elmts(FMAXN,7,-FMAXNPL:FMAXNTP,FMAXC)
      integer nbod,ntp,n,FBUFN,filter_D(FMAXC)

      integer i,j,k,id

!$OMP PARALLEL
!$OMP& SHARED (ntp, filter_D, FBUFN, elmts)
!$OMP DO
!$OMP& PRIVATE (i, j, k, id)
      do i=2,nbod
        id=-i
        do k=1,7
          do j=filter_D(n)+1,FBUFN
            elmts(j-filter_D(n),k,id,n)=elmts(j,k,id,n)
          enddo
        enddo
      enddo
!$OMP END DO NOWAIT

!$OMP DO
!$OMP& PRIVATE (i, j, k, id)
      do i = 1, ntp
        id=i
        do k=1,7
          do j=filter_D(n)+1,FBUFN
            elmts(j-filter_D(n),k,id,n)=elmts(j,k,id,n)
          enddo
        enddo
      enddo
!$OMP END DO
!$OMP END PARALLEL

      return
      end	! filter_shift

